from .COAsT import COAsT
from warnings import warn
import numpy as np
import xarray as xr


class DOMAIN(COAsT):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Get depths at time zero
        self.set_timezero_depth()

        return

    def set_dimension_mapping(self):
        self.dim_mapping = {'t':'t_dim', 'z':'z_dim',
                            'y':'y_dim', 'x':'x_dim'}
        #self.dim_mapping = None

    def construct_depths_from_spacings(self):
        # NEMO4 constucts depths from verical spacing variables
        self.depth_t = xr.DataArray( self.dataset.e3t_0.cumsum( dim='z_dim' ).squeeze() ) # size: nz,my,nx
        self.depth_t.attrs['units'] = 'm'
        self.depth_t.attrs['standard_name'] = 'depth_at_t-points'

        self.depth_w = xr.DataArray( self.dataset.e3w_0.cumsum( dim='z_dim' ).squeeze() ) # size: nz,my,nx
        self.depth_w.attrs['units'] = 'm'
        self.depth_w.attrs['standard_name'] = 'depth_at_w-points'

    def subset_indices(self, start: tuple, end: tuple, grid_ref: str = 'T') -> tuple:
        """
        based off transect_indices, this method looks to return all indices between the given points.
        This results in a 'box' (Quadrilateral) of indices.
        consequently the returned lists may have different lengths.

        :param start: A lat/lon pair
        :param end: A lat/lon pair
        :param grid_ref: The gphi/glam version a user wishes to search over
        :return: list of y indices, list of x indices,
        """
        assert isinstance(grid_ref, str) and grid_ref.upper() in ("T", "V", "U", "F"), \
            "grid_ref should be either \"T\", \"V\", \"U\", \"F\""

        letter = grid_ref.lower()

        [j1, i1] = self.find_j_i(start[0], start[1], letter)  # lat , lon
        [j2, i2] = self.find_j_i(end[0], end[1], letter)  # lat , lon

        return list(np.arange(j1, j2+1)), list(np.arange(i1, i2+1))





from .COAsT import COAsT
from scipy.ndimage import convolve1d
from scipy import interpolate
import gsw
import xarray as xr
import numpy as np


# =============================================================================
# The TRANSECT module is a place for code related to transects only
# =============================================================================

class Transect:
    
    def __init__(self, point_A: tuple, point_B: tuple, nemo_F: COAsT,
                 nemo_T: COAsT=None, nemo_U: COAsT=None, nemo_V: COAsT=None ):
        '''
        Class defining a generic transect type, which is a 3d dataset between a point A and 
        a point B, with a time dimension, a depth dimension and a transect dimension. The 
        transect dimension defines the points along the transect.
        The model Data is subsetted in its entirety along these dimensions.
        
        Note that Point A should be of lower latitude than point B.
        
        Example usage:
            point_A = (54,-15)
            point_B = (56,-12)
            transect = coast.Transect( point_A, point_B, nemo_t, nemo_u, nemo_v, nemo_f )
            

        Parameters
        ----------
        point_A : tuple, (lat,lon)
        point_B : tuple, (lat,lon)
        nemo_F : NEMO, model data on the F grid, which is required to define the 
                    transect
        nemo_T : NEMO, optional, model data on the T grid
        nemo_U : NEMO, optional, model data on the U grid
        nemo_V : NEMO, optional, model data on the V grid
        
        
        '''
        
        # point A should be of lower latitude than point B
        if abs(point_B[0]) < abs(point_A[0]):
            self.point_A = point_B
            self.point_B = point_A
        else:
            self.point_A = point_A
            self.point_B = point_B
            
        # Get points on transect
        tran_y_ind, tran_x_ind = self.get_transect_indices( nemo_F )
                
        # indices along the transect        
        self.y_ind = tran_y_ind
        self.x_ind = tran_x_ind
        self.len = len(tran_y_ind)
        self.data_tran = xr.Dataset()
        #self.normal_velocities = None
        #self.depth_integrated_transport_across_AB = None # (time,transect_segment_index)
        
        # Subset the nemo data along the transect creating a new dimension (r_dim),
        # which is a paramterisation for x_dim and y_dim defining the transect
        da_tran_y_ind = xr.DataArray( tran_y_ind, dims=['r_dim'])
        da_tran_x_ind = xr.DataArray( tran_x_ind, dims=['r_dim'])
        self.data_F = nemo_F.dataset.isel(y_dim = da_tran_y_ind, x_dim = da_tran_x_ind)
        if nemo_T is not None:
            self.data_T = nemo_T.dataset.isel(y_dim = da_tran_y_ind, x_dim = da_tran_x_ind)
        if nemo_U is not None:
            self.data_U = nemo_U.dataset.isel(y_dim = da_tran_y_ind, x_dim = da_tran_x_ind)
        if nemo_V is not None:
            self.data_V = nemo_V.dataset.isel(y_dim = da_tran_y_ind, x_dim = da_tran_x_ind)
        # For calculations we need access to a halo of points around the transect
        # self.data_n = dataset.isel(y=tran_y+1,x=tran_x)  
        # self.data_e = dataset.isel(y=tran_y,x=tran_x+1) 
        # self.data_s = dataset.isel(y=tran_y-1,x=tran_x) 
        # self.data_w = dataset.isel(y=tran_y,x=tran_x-1) 
        
        
 
    def get_transect_indices(self, nemo_F):
        '''
        Get the transect indices on a specific grid

        Parameters
        ----------
        nemo_F : the model grid to define the transect on
        
        Return
        ----------
        tran_y : array of y_dim indices
        tran_x : array of x_dim indices

        '''
        tran_y_ind, tran_x_ind, tran_len = nemo_F.transect_indices(self.point_A, self.point_B)
        tran_y_ind = np.asarray(tran_y_ind)
        tran_x_ind = np.asarray(tran_x_ind)
        
        # Redefine transect so that each point on the transect is seperated
        # from its neighbours by a single index change in y or x, but not both
        dist_option_1 = nemo_F.dataset.e2.values[tran_y_ind, tran_x_ind] + nemo_F.dataset.e1.values[tran_y_ind+1, tran_x_ind]
        dist_option_2 = nemo_F.dataset.e2.values[tran_y_ind, tran_x_ind+1] + nemo_F.dataset.e1.values[tran_y_ind, tran_x_ind]
        spacing = np.abs( np.diff(tran_y_ind) ) + np.abs( np.diff(tran_x_ind) )
        spacing[spacing!=2]=0
        doublespacing = np.nonzero( spacing )[0]
        for ispacing in doublespacing[::-1]:
            if dist_option_1[ispacing] < dist_option_2[ispacing]:
                tran_y_ind = np.insert( tran_y_ind, ispacing+1, tran_y_ind[ispacing+1] )
                tran_x_ind = np.insert( tran_x_ind, ispacing+1, tran_x_ind[ispacing] )
            else:
                tran_y_ind = np.insert( tran_y_ind, ispacing+1, tran_y_ind[ispacing] )
                tran_x_ind = np.insert( tran_x_ind, ispacing+1, tran_x_ind[ispacing+1] ) 
        return (tran_y_ind, tran_x_ind)
        
    
    def transport_across_AB(self):
        """
    
        Computes the flow through the transect at each segment and stores:
        Transect normal velocities at each grid point in Transect.normal_velocities,
        Depth integrated volume transport across the transect at each transect segment in 
        Transect.depth_integrated_transport_across_AB
        
        Return 
        -----------
        Transect normal velocities at each grid point (m/s)
        Depth integrated volume transport across the transect at each transect segment (Sv)
        """
        
        velocity = np.ma.zeros(np.shape(self.data_U.vozocrtx))
        vol_transport = np.ma.zeros(np.shape(self.data_U.vozocrtx))
        depth_integrated_transport = np.ma.zeros( np.shape(self.data_U.vozocrtx[:,0,:] )) 
        depth_0 = np.ma.zeros(np.shape(self.data_U.depth_0))
        
        dy = np.diff(self.y_ind)
        dx = np.diff(self.x_ind)
        # Loop through each point along the transact
        for idx in np.arange(0, self.len-1):            
            if dy[idx] > 0:
                # u flux (+ in)
                velocity[:,:,idx] = self.data_U.vozocrtx[:,:,idx+1].to_masked_array()
                vol_transport[:,:,idx] = ( velocity[:,:,idx] * self.data_U.e2[idx+1].to_masked_array() *
                                          self.data_U.e3[:,:,idx+1].to_masked_array() )
                depth_integrated_transport[:,idx] = np.sum( vol_transport[:,:,idx], axis=1 )
                depth_0[:,idx] = self.data_U.depth_0[:,idx+1].to_masked_array()
            elif dx[idx] > 0:
                # v flux (- in) 
                velocity[:,:,idx] = - self.data_V.vomecrty[:,:,idx+1].to_masked_array()
                vol_transport[:,:,idx] = ( velocity[:,:,idx] * self.data_V.e1[idx+1].to_masked_array() *
                                          self.data_V.e3[:,:,idx+1].to_masked_array() )
                depth_integrated_transport[:,idx] = np.sum( vol_transport[:,:,idx], axis=1 )
                depth_0[:,idx] = self.data_V.depth_0[:,idx+1].to_masked_array()
            elif dx[idx] < 0:
                # v flux (+ in)
                velocity[:,:,idx] = self.data_V.vomecrty[:,:,idx].to_masked_array()
                vol_transport[:,:,idx] = ( velocity[:,:,idx] * self.data_V.e1[idx].to_masked_array() *
                                          self.data_V.e3[:,:,idx].to_masked_array() )
                depth_integrated_transport[:,idx] = np.sum( vol_transport[:,:,idx], axis=1 )
                depth_0[:,idx] = self.data_V.depth_0[:,idx].to_masked_array()
        
        
        dimensions = ['t_dim', 'z_dim', 'r_dim']
        
  #      self.data_tran['depth_0'] = xr.DataArray( depth_0[:,:-1], dims=['z_dim', 'r_dim'], 
   #                attrs={'Units':'m', 'standard_name': 'Initial depth at time zero',
    #                'long_name': 'Initial depth at time zero defined at the normal velocity grid points on the transect'})
 
   
        self.data_tran['normal_velocities'] = xr.DataArray( velocity[:,:,:-1], 
                    coords={'time': (('t_dim'), self.data_U.time.values),'depth_0': (('z_dim','r_dim'), depth_0[:,:-1])},
                    dims=['t_dim', 'z_dim', 'r_dim'] )        
    
        self.data_tran['depth_integrated_transport_across_AB'] = xr.DataArray( depth_integrated_transport[:,:-1] / 1000000.,
                    coords={'time': (('t_dim'), self.data_U.time.values)},
                    dims=['t_dim', 'r_dim'] ) 
        
        self.data_tran.depth_0.attrs['units'] = 'm'
        self.data_tran.depth_0.attrs['standard_name'] = 'Initial depth at time zero'
        self.data_tran.depth_0.attrs['long_name'] = 'Initial depth at time zero defined at the normal velocity grid points on the transect'
                        
        return  
    

    def moving_average(self, array_to_smooth, window=2, axis=-1):
        '''
        Returns the input array smoothed along the given axis using convolusion
        '''
        return convolve1d( array_to_smooth, np.ones(window), axis=axis ) / window
    
    def interpolate_slice(self, variable_slice, depth, interpolated_depth=None ):
        '''
        Linearly interpolates the variable at a single time along the z_dim, which must be the
        first axis.

        Parameters
        ----------
        variable_slice : Variable to interpolate (z_dim, transect_dim)
        depth : The depth at each z point for each point along the transect
        interpolated_depth : (optional) desired depth profile to interpolate to. If not supplied
            a uniform depth profile uniformaly spaced between zero and variable max depth will be used
            with a spacing of 2 metres.

        Returns
        -------
        interpolated_depth_variable_slice : Interpolated variable
        interpolated_depth : Interpolation depth

        '''
        if interpolated_depth is None:
            interpolated_depth = np.arange(0, np.nanmax(depth), 2)
            
        interpolated_depth_variable_slice = np.zeros( (len(interpolated_depth), variable_slice.shape[-1]) )
        for i in np.arange(0, variable_slice.shape[-1] ):
            depth_func = interpolate.interp1d( depth[:,i], variable_slice[:,i], axis=0, bounds_error=False )        
            interpolated_depth_variable_slice[:,i] = depth_func( interpolated_depth )
            
        return (interpolated_depth_variable_slice, interpolated_depth )
    
    
    def plot_normal_velocity(self, time, plot_info: dict, cmap, smoothing_window=0):  
        '''
            Quick plot routine of velocity across the transect AB at a specific time.
            An option is provided to smooth the velocities along the transect.
            NOTE: For smoothing use even integers to smooth the x and y velocities together
    
    
    
        Parameters
        ---------------
        time: either as integer index or actual time as a string.
        plot_info: dictionary of infomation {'fig_size': value, 'title': value, 'vmin':value, 'vmax':value}
        Note that if vmin and max are not set then the colourbar will be centred at zero
        smoothing_window: smoothing via convolusion, larger number applies greater smoothing, recommended
        to use even integers

        
        '''
        try:
            data = self.data_tran.sel(t_dim = time)
        except KeyError:
            data = self.data_tran.isel(t_dim = time)        
        
        if smoothing_window != 0:
            normal_velocities, depth = self.interpolate_slice( data.normal_velocities, data.depth_0 )            
            normal_velocities = self.moving_average(normal_velocities, smoothing_window, axis=-1)
            r_dim_2d = np.broadcast_to( data.r_dim, normal_velocities.shape  )
        else:
            normal_velocities = data.normal_velocities
            depth = data.depth_0
            _ , r_dim_2d = xr.broadcast( depth, data.r_dim  )
                    
        import matplotlib.pyplot as plt
        plt.close('all')
        fig = plt.figure(figsize=plot_info['fig_size'])
        ax = fig.gca()

        plt.pcolormesh(r_dim_2d, depth, normal_velocities, cmap=cmap)
            
        plt.title(plot_info['title'])
        plt.ylabel('Depth [m]')
        try:
            plt.clim(vmin=plot_info['vmin'], vmax=plot_info['vmax'])
        except KeyError:
            lim = np.nanmax(np.abs(normal_velocities))
            plt.clim(vmin=-lim, vmax=lim)
        plt.xticks([0,data.r_dim.values[-1]],['A','B'])
        plt.colorbar(label='Velocities across AB [m/s]')
        plt.gca().invert_yaxis()

        plt.show()
        return fig,ax
        
    
    def plot_depth_integrated_transport(self, time, plot_info: dict, smoothing_window=0):
        '''
            Quick plot routine of depth integrated transport across the transect AB at a specific time.
            An option is provided to smooth along the transect via convolution, 
            NOTE: For smoothing use even integers to smooth the x and y velocities together
    
        Parameters
        ---------------
        time: either as integer index or actual time as a string.
        plot_info: dictionary of infomation {'fig_size': value, 'title': value}
        smoothing_window: smoothing via convolusion, larger number applies greater smoothing. 
        Recommended to use even integers.
        returns: pyplot object
        '''
        try:
            data = self.data_tran.sel(t_dim = time)
        except KeyError:
            data = self.data_tran.isel(t_dim = time)            
        
        if smoothing_window != 0:    
            transport = self.moving_average(data.depth_integrated_transport_across_AB, smoothing_window, axis=-1)
        else:
            transport = data.depth_integrated_transport_across_AB
        
        import matplotlib.pyplot as plt
        plt.close('all')
        fig = plt.figure(figsize=plot_info['fig_size'])
        ax = fig.gca()

        plt.plot( data.r_dim, transport )

        plt.title(plot_info['title'])
        plt.xticks([0,data.r_dim[-1]],['A','B'])
        plt.ylabel('Volume transport across AB [SV]')
        plt.show()
        return fig,ax
    
    
    def construct_density_on_z_levels( self, EOS='EOS10'):#, z_levels=None ):        
        '''
            For s-level model output this method recontructs the in-situ density 
            onto z_levels along the transect. The z_levels and density field
            are added to the data_T dataset attribute. 
            
            Requirements: The supplied t-grid dataset must contain the 
            Practical Salinity and the Potential Temperature variables. The depth_0
            field must also be supplied. The GSW package is used to calculate
            The Absolute Pressure, Absolute Salinity and Conservate Temperature.
            
            This method is useful when horizontal gradients of the density field 
            are required. Currently z_levels cannot be specified, the 
            method will contruct a z_levels profile from the s_levels.
            
            Note that currently density can only be constructed using the EOS10
            equation of state.

        Parameters
        ----------
        EOS : equation of state, optional
            DESCRIPTION. The default is 'EOS10'.


        Returns
        -------
        None.
        adds attributes Transect.data_T.depth_z_levels and Transect.data_T.density_z_levels

        '''        

        
        try:
            if EOS != 'EOS10': 
                raise ValueError(str(self) + ': Density calculation for ' + EOS + ' not implemented.')
            if self.data_T is None:
                raise ValueError(str(self) + ': Density calculation can only be performed for a t-grid object,\
                                 the tracer grid for NEMO.' )
          #  if not self.data_T.ln_sco.item():
           #     raise ValueError(str(self) + ': Density calculation only implemented for s-vertical-coordinates.')            
    
            #if z_levels is None:
            z_levels = self.data_T.depth_0.max(dim=(['r_dim']))                
            z_levels_min = self.data_T.depth_0[0,:].max(dim=(['r_dim']))
            z_levels[0] = z_levels_min
            #else:
             #   z_max = self.data_T.depth_0.max(dim=(['r_dim','z_dim'])).item()
              #  z_min = self.data_T.depth_0[0,:].max(dim=(['r_dim'])).item()
               # z_levels = z_levels[z_levels<=z_max]
                #z_levels = z_levels[z_levels>=z_min]
                
                
            
            try:    
                shape_ds = ( self.data_T.t_dim.size, z_levels.size, 
                                self.data_T.r_dim.size )
                sal = self.data_T.salinity.to_masked_array()
                temp = self.data_T.temperature.to_masked_array()                
            except AttributeError:
                shape_ds = ( 1, z_levels.size, self.data_T.r_dim.size )
                sal = self.data_T.salinity.to_masked_array()[np.newaxis,...]
                temp = self.data_T.temperature.to_masked_array()[np.newaxis,...]
            
            sal_z_levels = np.ma.zeros( shape_ds )
            temp_z_levels = np.ma.zeros( shape_ds )
            density_z_levels = np.ma.zeros( shape_ds )
            
            s_levels = self.data_T.depth_0.to_masked_array()
            lat = self.data_T.latitude.values
            lon = self.data_T.longitude.values
    
            for it in np.arange(0, shape_ds[0]):
                for ir in self.data_T.r_dim:
                    if np.all(np.isnan(sal[it,:,ir])):
                        density_z_levels[it,:,ir] = np.nan
                        density_z_levels[it,:,ir].mask = True
                    else:                      
                        sal_func = interpolate.interp1d( s_levels[:,ir], sal[it,:,ir], 
                                    bounds_error=False, kind='linear')
                        temp_func = interpolate.interp1d( s_levels[:,ir], temp[it,:,ir], 
                                    bounds_error=False, kind='linear')
                        
                        sal_z_levels[it,:,ir] = sal_func(z_levels.values)
                        temp_z_levels[it,:,ir] = temp_func(z_levels.values)
                        
            # Absolute Pressure    
            pressure_absolute = np.ma.masked_invalid(
                gsw.p_from_z( -z_levels.values[:,np.newaxis], lat ) ) # depth must be negative           
            # Absolute Salinity           
            sal_absolute = np.ma.masked_invalid(
                gsw.SA_from_SP( sal_z_levels, pressure_absolute, lon, lat ) )
            sal_absolute = np.ma.masked_less(sal_absolute,0)
            # Conservative Temperature
            temp_conservative = np.ma.masked_invalid(
                gsw.CT_from_pt( sal_absolute, temp_z_levels ) )
            # In-situ density
            density_z_levels = np.ma.masked_invalid( gsw.rho( 
                sal_absolute, temp_conservative, pressure_absolute ) )
            
            coords={'depth_z_levels': (('z_dim'), z_levels.values),
                    'latitude': (('r_dim'), self.data_T.latitude.values),
                    'longitude': (('r_dim'), self.data_T.longitude.values)}
            dims=['z_dim', 'r_dim']
            attributes = {'units': 'kg / m^3', 'standard name': 'In-situ density on the z-level vertical grid'}
            
            if shape_ds[0] != 1:
                coords['time'] = (('t_dim'), self.data_T.time.values)
                dims.insert(0, 't_dim')
              
            self.data_T['density_z_levels'] = xr.DataArray( np.squeeze(density_z_levels), 
                    coords=coords, dims=dims, attrs=attributes )

        except AttributeError as err:
            print(err)
            
        return
        
        



from .COAsT import COAsT
import numpy as np
import xarray as xr
from warnings import warn
import copy
import gsw
#from .CRPS import CRPS
#from .interpolate_along_dimension import interpolate_along_dimension

class INTERNALTIDE(COAsT):
    """
    Object for handling and storing necessary information, methods and outputs
    for calculation of internal tide diagnostics.

    Herein the depth moments of stratification are used as proxies for
    pycnocline depth (as the first  moment of stratification), and pycnocline
    thickness  (as the 2nd moment of stratification).
    This approximation improves towards the limit of a two-layer fluid.

    For stratification that is not nearly two-layer, the pycnocline
    thickness appears large and this method for identifying the pycnocline
    depth is less reliable.

    Parameters
    ----------
        nemo_t : xr.Dataset
            nemo object on t-points.
        nemo_w : xr.Dataset, optional
            nemo object on w-points.

    Example basic usage:
    -------------------
        # Create Internal tide diagnostics object
        IT_obj = INTERNALTIDE(nemo_t, nemo_w) # For NEMO objects on t and w-pts
        IT_obj.construct_pycnocline_vars( nemo_t, nemo_w )
        # Make maps of pycnocline thickness and depth
        IT_obj.quick_plot()
    """
    def __init__(self, nemo_t: xr.Dataset, nemo_w: xr.Dataset):

        self.dataset = xr.Dataset()

        # Define the spatial dimensional size and check the dataset and domain arrays are the same size in z_dim, ydim, xdim
        self.nt = nemo_t.dataset.dims['t_dim']
        self.nz = nemo_t.dataset.dims['z_dim']
        self.ny = nemo_t.dataset.dims['y_dim']
        self.nx = nemo_t.dataset.dims['x_dim']


    def construct_pycnocline_vars( self, nemo_t: xr.Dataset, nemo_w: xr.Dataset, strat_thres = -0.01):
        """
        Computes depth moments of stratifcation. Under the assumption that the
        stratification approximately represents a two-layer fluid, these can be
        interpreted as pycnocline depths and thicknesses. They are computed on
        w-points.

        1st moment of stratification: \int z.strat dz / \int strat dz
            In the limit of a two layer fluid this is equivalent to the
        pycnocline depth, or z_d (units: metres)

        2nd moment of stratification: \sqrt{\int (z-z_d)^2 strat dz / \int strat dz}
            where strat = d(density)/dz
            In the limit of a two layer fluid this is equivatlent to the
        pycnocline thickness, or z_t (units: metres)

        Parameters
        ----------
        nemo_t : xr.Dataset
            nemo object on t-points.
        nemo_w : xr.Dataset, optional
            nemo object on w-points.
        strat_thres: float - Optional
            limiting stratification (rho_dz < 0) to trigger masking of mixed waters

        Output
        ------
        self.dataset.strat_1st_mom - (t,y,x) pycnocline depth
        self.dataset.strat_2nd_mom - (t,y,x) pycnocline thickness
        self.dataset.strat_1st_mom_masked - (t,y,x) pycnocline depth, masked
                in weakly stratified water beyond strat_thres
        self.dataset.strat_2nd_mom_masked - (t,y,x) pycnocline thickness, masked
                in weakly stratified water beyond strat_thres
        self.dataset.mask - (t,y,x) [1/0] stratified/unstrafied
                water column according to strat_thres not being met anywhere
                in the column

        Returns
        -------
        None.

        Example Usage
        -------------
        # load some example data
        dn_files = "./example_files/"
        dn_fig = 'unit_testing/figures/'
        fn_nemo_grid_t_dat = 'nemo_data_T_grid_Aug2015.nc'
        fn_nemo_dom = 'COAsT_example_NEMO_domain.nc'
        nemo_t = coast.NEMO(dn_files + fn_nemo_grid_t_dat,
                     dn_files + fn_nemo_dom, grid_ref='t-grid')
        # create an empty w-grid object, to store stratification
        nemo_w = coast.NEMO( fn_domain = dn_files + fn_nemo_dom,
                           grid_ref='w-grid')

        # initialise Internal Tide object
        IT = coast.INTERNALTIDE(nemo_t, nemo_w)
        # Construct pycnocline variables: depth and thickness
        IT.construct_pycnocline_vars( nemo_t, nemo_w )
        # Plot pycnocline depth and thickness
        IT.quickplot()

        """
        #%% Contruct in-situ density if not already done
        if not hasattr(nemo_t.dataset, 'density'):
            nemo_t.construct_density( EOS='EOS10' )

        #%% Construct stratification if not already done. t-pts --> w-pts
        if not hasattr(nemo_w.dataset, 'rho_dz'):
            nemo_w = nemo_t.differentiate( 'density', dim='z_dim', out_varstr='rho_dz', out_obj=nemo_w ) # --> sci_nwes_w.rho_dz

        # Define the spatial dimensional size and check the dataset and domain arrays are the same size in z_dim, ydim, xdim
        nt = nemo_t.dataset.dims['t_dim']
        nz = nemo_t.dataset.dims['z_dim']
        ny = nemo_t.dataset.dims['y_dim']
        nx = nemo_t.dataset.dims['x_dim']

        ## Create a mask for weakly stratified waters
        # Preprocess stratification
        strat = copy.copy(nemo_w.dataset.rho_dz)  # (t_dim, z_dim, ydim, xdim). w-pts.
        # Ensure surface value is 0
        strat[:,0,:,:] = 0
        # Ensure bed value is 0
        strat[:,-1,:,:] = 0
        # mask out the Nan values
        strat = strat.where( ~xr.ufuncs.isnan(nemo_w.dataset.rho_dz), drop=False )
        # create mask with a stratification threshold
        strat_m = nemo_w.dataset.latitude * 0 + 1 # create a stratification mask: [1/0] = strat/un-strat
        strat_m = strat_m.where( strat.min(dim='z_dim').squeeze() < strat_thres, 0, drop=False)
        strat_m = strat_m.transpose("t_dim", "y_dim", "x_dim", transpose_coords=True)

        ## Compute statification variables
        # initialise pycnocline variables
        zd = np.zeros((nt, ny, nx)) # pycnocline depth
        zt = np.zeros((nt, ny, nx)) # pycnocline thickness

        ## Construct intermediate variables
        # Broadcast to fill out missing (time) dimensions in grid data
        _, depth_0_4d = xr.broadcast(strat, nemo_w.dataset.depth_0)
        _, e3_0_4d    = xr.broadcast(strat, nemo_w.dataset.e3_0.squeeze())

        # intergrate strat over depth
        intN2  = ( strat * e3_0_4d ).sum( dim='z_dim', skipna=True)
        # intergrate (depth * strat) over depth
        intzN2 = (strat * e3_0_4d * depth_0_4d).sum( dim='z_dim', skipna=True)

        # compute pycnocline depth
        zd = (intzN2 / intN2 )# pycnocline depth

        # compute pycnocline thickness
        intz2N2 = ( xr.ufuncs.square(depth_0_4d - zd) * e3_0_4d * strat  ).sum( dim='z_dim', skipna=True )
        zt = ( xr.ufuncs.sqrt(intz2N2 / intN2) ) # pycnocline thickness

        # Define xarray attributes
        coords = {'time': (('t_dim'), nemo_t.dataset.time.values),
                    'latitude': (('y_dim','x_dim'), nemo_t.dataset.latitude.values),
                    'longitude': (('y_dim','x_dim'), nemo_t.dataset.longitude.values)}
        dims = ['t_dim', 'y_dim', 'x_dim']

        # Save a xarray objects
        self.dataset['strat_2nd_mom'] = xr.DataArray( zt,
                    coords=coords, dims=dims)
        self.dataset.strat_2nd_mom.attrs['units'] = 'm'
        self.dataset.strat_2nd_mom.attrs['standard_name'] = 'pycnocline thickness'
        self.dataset.strat_2nd_mom.attrs['long_name'] = 'Second depth moment of stratification'


        self.dataset['strat_1st_mom'] = xr.DataArray( zd,
                    coords=coords, dims=dims )
        self.dataset.strat_1st_mom.attrs['units'] = 'm'
        self.dataset.strat_1st_mom.attrs['standard_name'] = 'pycnocline depth'
        self.dataset.strat_1st_mom.attrs['long_name'] = 'First depth moment of stratification'


        #%% Mask pycnocline variables in weak stratification
        #zd_m = zd.where( strat_m > 0, -999, drop=False )
        #zt_m = zt.where( strat_m > 0, -999, drop=False )
        zd_m = zd.where( strat_m > 0 )
        zt_m = zt.where( strat_m > 0 )

        self.dataset['mask'] = xr.DataArray( strat_m,
                    coords=coords, dims=dims)

        self.dataset['strat_2nd_mom_masked'] = xr.DataArray( zt_m,
                    coords=coords, dims=dims)
        self.dataset.strat_2nd_mom_masked.attrs['units'] = 'm'
        self.dataset.strat_2nd_mom_masked.attrs['standard_name'] = 'masked pycnocline thickness'
        self.dataset.strat_2nd_mom_masked.attrs['long_name'] = 'Second depth moment of stratification, masked in weak stratification'


        self.dataset['strat_1st_mom_masked'] = xr.DataArray( zd_m,
                    coords=coords, dims=dims )
        self.dataset.strat_1st_mom_masked.attrs['units'] = 'm'
        self.dataset.strat_1st_mom_masked.attrs['standard_name'] = 'masked pycnocline depth'
        self.dataset.strat_1st_mom_masked.attrs['long_name'] = 'First depth moment of stratification, masked in weak stratification'

    def quick_plot(self, var : xr.DataArray = None):
        """

        Map plot for pycnocline depth and thickness variables.

        Parameters
        ----------
        var : xr.DataArray, optional
            Pass variable to plot. The default is None. In which case both
            strat_1st_mom and strat_2nd_mom are plotted.

        Returns
        -------
        None.

        Example Usage
        -------------
        IT.quick_plot( 'strat_1st_mom_masked' )

        """
        import matplotlib.pyplot as plt

        if var is None:
            var_lst = [self.dataset.strat_1st_mom_masked, self.dataset.strat_2nd_mom_masked]
        else:
            var_lst = [self.dataset[var]]

        for var in var_lst:
            fig = plt.figure(figsize=(10, 10))
            ax = fig.gca()
            plt.pcolormesh( self.dataset.longitude.squeeze(),
                           self.dataset.latitude.squeeze(),
                           var.isel(t_dim = 0) )
            #               var.mean(dim = 't_dim') )
            #plt.contourf( self.dataset.longitude.squeeze(),
            #               self.dataset.latitude.squeeze(),
            #               var.mean(dim = 't_dim'), levels=(0,10,20,30,40) )
            title_str = self.dataset.time[0].dt.strftime("%d %b %Y: ").values \
                + var.attrs['standard_name'] +  ' (' + var.attrs['units'] + ')'
            plt.title(title_str)
            plt.xlabel('longitude')
            plt.ylabel('latitude')
            plt.clim([0, 50])
            plt.colorbar()
            plt.show()
        return fig,ax



from .COAsT import COAsT
from warnings import warn
import numpy as np
import xarray as xr

class OBSERVATION(COAsT):

    def set_dimension_mapping(self):
        self.dim_mapping = None

    def subset_indices_lonlat_box(self, lonbounds, latbounds):
        """Generates array indices for data which lies in a given lon/lat box.

        Keyword arguments:
        lon       -- Longitudes, 1D or 2D.
        lat       -- Latitudes, 1D or 2D
        lonbounds -- Array of form [min_longitude=-180, max_longitude=180]
        latbounds -- Array of form [min_latitude, max_latitude]
        
        return: Indices corresponding to datapoints inside specified box
        """
        lon = self.dataset.longitude.copy()
        lat = self.dataset.latitude
        lon[lon>180] = lon[lon>180] - 360
        lon[lon<-180] = lon[lon<-180] + 360
        ff1 = ( lon > lonbounds[0] ).astype(int)
        ff2 = ( lon < lonbounds[1] ).astype(int)
        ff3 = ( lat > latbounds[0] ).astype(int)
        ff4 = ( lat < latbounds[1] ).astype(int)
        indices = np.where( ff1 * ff2 * ff3 * ff4 )
        return indices[0]
    
    def adjust_longitudes(self, lonbounds=[-180,180]):
        bool0 = self['longitude']<lonbounds[0]
        bool1 = self['longitude']>lonbounds[1]
        self['longitude'][bool0] = self['longitude'][bool0] + 360
        self['longitude'][bool1] = self['longitude'][bool1] - 360
        
    def interpolate_model_to_obs(self):
        print('Method not implemented for observation object type.')
        
    def quick_plot(self):
        print('Method not implemented for observation object type.')



import xarray as xr
import numpy as np
from .COAsT import COAsT

class interpolate_along_dimension():
    '''
    An object for flexibly and quickly interpolating an xarray variable along
    a specified dimension.
    '''
    
    def __init__(self, data, dim_interp, dim_name, method='nearest'):
        self.method = method
        self.dim_name = dim_name
        self.data = data
        self.weights = self.calculate_weights(self.data[dim_name], 
                                              dim_interp)
        
    def __getitem__(self, indices):
        if self.method == 'nearest':
            return self.get_nearest(self.data, indices)
        else:
            raise NotImplementedError
    
    def get_nearest(self, data, indices):
        ''' Returns the nearest interpolated data for specified indices '''
        return data.isel({self.dim_name : self.weights[indices]})
    
    def get_linear(self):
        ''' Returns the linearly interpolated data for specified indices '''
        raise NotImplementedError
        return
    
    def calculate_weights(self, x0, xi):
        ''' Control routine for sending data to different weights functions '''
        if self.method == 'nearest':
            weights = self.calculate_weights_nearest(x0, xi)
        else:
            raise NotImplementedError
        return weights
        
    def calculate_weights_nearest(self, x0, xi):
        ''' Calculates weights for nearest neighbour. In this case weights are
        just the indices of the data that is closest. '''
        vdiff = self.difference_matrix(x0, xi)
        return np.argmin(vdiff, axis=0)
        
    def calculate_weights_linear(self, x0, xi):
        ''' Calculates weights for linear interpolation. The form of each row 
        of the weights array is [w1, w2, i1, i2]. w1 and w2 are the weights and
        i1, i2 are the indices of the data that the weights are applied to'''
        raise NotImplementedError
        return

    def difference_matrix(self, v1, v2):
        ''' Calculates all pairwise absolute distances between two vectors.'''
        mg_x, mg_y = np.meshgrid(v1,v2)
        return np.transpose( np.abs(mg_x - mg_y) )


from dask import delayed
from dask import array
import xarray as xr
import numpy as np
from dask.distributed import Client
from warnings import warn
import copy
import scipy as sp

def setup_dask_clinet(workers=2, threads=2, memory_limit_per_worker='2GB'):
    Client(n_workers=workers, threads_per_worker=threads, memory_limit=memory_limit_per_worker)


class COAsT:
    def __init__(self, file = None, chunks: dict=None, multiple=False,
                 workers=2, threads=2, memory_limit_per_worker='2GB'):
        # self.client = Client(n_workers=workers, threads_per_worker=threads, memory_limit=memory_limit_per_worker)
        self.dataset = None
        self.earth_raids = 6371.007176 # Radius of the earth in km
        self.set_dimension_mapping()
        self.set_variable_mapping()
        #self.set_grid_ref_attr()
        if file is None:
            print("Object created but no file or directory specified: " + 
                  str(self) + " \n" +
                  "Use COAsT.load() to load a netCDF file from file path or " +
                  "directory into this object.")
        else:
            self.load(file, chunks, multiple)

    def load(self, file_or_dir, chunks:dict=None, multiple=False):
        """
        Loads a file into a COAsT object's dataset variable using xarray
        
        Args:
            file (str)     : file name or directory to multiple files.
            chunks (dict)  : Chunks to use in Dask [default None]
            multiple (bool): If true, load in multiple files from directory.
                             If false load a single file [default False]
        """
        if multiple:
            self.load_multiple(file_or_dir, chunks)
        else:
            self.load_single(file_or_dir, chunks)    

    def __getitem__(self, name: str):
        return self.dataset[name]

    def load_single(self, file, chunks: dict = None):
        """ Loads a single file into COAsT object's dataset variable. """
        self.dataset = xr.open_dataset(file, chunks=chunks)

    def load_multiple(self, directory_to_files, chunks: dict = None):
        """ Loads multiple files from directory into dataset variable. """
        self.dataset = xr.open_mfdataset(
            directory_to_files, chunks=chunks, parallel=True, 
            combine="by_coords") #, compat='override')

    def load_dataset(self, dataset):
        """        
        :param dataset: The dataset to use
        :type dataset: xarray.Dataset
        """
        
        self.dataset = dataset
    
    def set_dimension_mapping(self):
        self.dim_mapping = None
        
    def set_variable_mapping(self):
        self.var_mapping = None

    def set_grid_ref_attr(self):
        self.grid_ref_attr_mapping = None

    def set_dimension_names(self, dim_mapping: dict):
        """ 
        Relabel dimensions in COAsT object xarray.dataset to ensure
        consistent naming throughout the COAsT package.
        
        Args:
            dim_mapping (dict): keys are dimension names to change and values
                                new dimension names
        """
        if dim_mapping is None: return
        for key, value in dim_mapping.items():
            try:
                self.dataset = self.dataset.rename_dims({ key : value })
            except:
                pass
                #print(str(self) + ': Problem renaming dimension: ' + 
                #      key + ' -> ' + value)
                
    def set_variable_names(self, var_mapping: dict):
        """ 
        Relabel variables in COAsT object xarray.dataset to ensure
        consistent naming throughout the COAsT package.
        
        Args:
            var_mapping (dict): keys are variable names to change and values
                                are new variable names
        """
        if var_mapping is None: return
        for key, value in var_mapping.items():
            try:
                self.dataset = self.dataset.rename_vars({ key : value })
            except:
                pass
                #print(str(self) + ': Problem renaming variable: ' + 
                #      key + ' -> ' + value)

    def set_variable_grid_ref_attribute(self, grid_ref_attr_mapping: dict):
        """
        Set attributes for variables to access within package.
        Set grid attributes to identify with grid variable is associated with.
        """
        if grid_ref_attr_mapping is None: return
        for key, value in grid_ref_attr_mapping.items():
            try:
                self.dataset[key].attrs['grid_ref'] = value
            except:
                pass
                #print(str(self) + ': Problem assigning ' + key +
                #        'grid_ref attribute ' + value)

    def copy(self):
        return copy.copy(self)
        
    def isel(self, indexers: dict = None, drop: bool = False,
             **kwargs):
        '''
        Indexes COAsT object along specified dimensions using xarray isel. 
        Input is of same form as xarray.isel. Basic use, hand in either:
            1. Dictionary with keys = dimensions, values = indices
            2. **kwargs of form dimension = indices
        '''
        obj_copy = self.copy()
        obj_copy.dataset = obj_copy.dataset.isel(indexers, drop, **kwargs)
        return obj_copy
    
    def sel(self, indexers: dict = None, drop: bool = False,
             **kwargs):
        '''
        Indexes COAsT object along specified dimensions using xarray sel. 
        Input is of same form as xarray.sel. Basic use, hand in either:
            1. Dictionary with keys = dimensions, values = indices
            2. **kwargs of form dimension = indices
        '''
        obj_copy = self.copy()
        obj_copy.dataset = obj_copy.dataset.sel(indexers, drop, **kwargs)
        return obj_copy
    
    def rename(self, rename_dict, inplace: bool=None, **kwargs):
        self.dataset = self.dataset.rename(rename_dict, inplace, **kwargs)
        return

    def subset(self, **kwargs):
        '''
        Subsets all variables within the dataset inside self (a COAsT object).
        Input is a set of keyword argument pairs of the form:
            dimension_name = indices
        The entire object is then subsetted along this dimension at indices
        '''
        self.dataset = self.dataset.isel(kwargs)
        
    def subset_as_copy(self, **kwargs):
        '''
        Similar to COAsT.subset() however applies the subsetting to a copy of
        the original COAsT object. This subsetted copy is then returned.
        Useful for preserving the original object whilst creating smaller
        subsetted object copies.
        '''
        obj_copy = self.copy()
        obj_copy.subset(**kwargs)
        return obj_copy

    def distance_between_two_points(self):
        raise NotImplementedError
        
    def subset_indices_by_distance(self, centre_lon: float, centre_lat: float, 
                                   radius: float):
        """
        This method returns a `tuple` of indices within the `radius` of the lon/lat point given by the user.

        Distance is calculated as haversine - see `self.calculate_haversine_distance`

        :param centre_lon: The longitude of the users central point
        :param centre_lat: The latitude of the users central point
        :param radius: The haversine distance (in km) from the central point
        :return: All indices in a `tuple` with the haversine distance of the central point
        """

        # Flatten NEMO domain stuff.
        lon = self.dataset.longitude
        lat = self.dataset.latitude

        # Calculate the distances between every model point and the specified
        # centre. Calls another routine dist_haversine.

        dist = self.calculate_haversine_distance(centre_lon, centre_lat, 
                                                 lon, lat)
        indices_bool = dist < radius
        indices = np.where(indices_bool.compute())

        return xr.DataArray(indices[0]), xr.DataArray(indices[1])
    
    def subset_indices_lonlat_box(self, lonbounds, latbounds):
        """Generates array indices for data which lies in a given lon/lat box.

        Keyword arguments:
        lon       -- Longitudes, 1D or 2D.
        lat       -- Latitudes, 1D or 2D
        lonbounds -- Array of form [min_longitude=-180, max_longitude=180]
        latbounds -- Array of form [min_latitude, max_latitude]
        
        return: Indices corresponding to datapoints inside specified box
        """
        lon_str = 'longitude'
        lat_str = 'latitude'
        lon = self.dataset[lon_str].copy()
        lat = self.dataset[lat_str]
        ff = ( lon > lonbounds[0] ).astype(int)
        ff = ff*( lon < lonbounds[1] ).astype(int)
        ff = ff*( lat > latbounds[0] ).astype(int)
        ff = ff*( lat < latbounds[1] ).astype(int)
        return np.where(ff)

    def calculate_haversine_distance(self, lon1, lat1, lon2, lat2):
        '''
        # Estimation of geographical distance using the Haversine function.
        # Input can be single values or 1D arrays of locations. This
        # does NOT create a distance matrix but outputs another 1D array.
        # This works for either location vectors of equal length OR a single loc
        # and an arbitrary length location vector.
        #
        # lon1, lat1 :: Location(s) 1.
        # lon2, lat2 :: Location(s) 2.
        '''

        # Convert to radians for calculations
        lon1 = xr.ufuncs.deg2rad(lon1)
        lat1 = xr.ufuncs.deg2rad(lat1)
        lon2 = xr.ufuncs.deg2rad(lon2)
        lat2 = xr.ufuncs.deg2rad(lat2)

        # Latitude and longitude differences
        dlat = (lat2 - lat1) / 2
        dlon = (lon2 - lon1) / 2

        # Haversine function.
        distance = xr.ufuncs.sin(dlat) ** 2 + xr.ufuncs.cos(lat1) * xr.ufuncs.cos(lat2) * xr.ufuncs.sin(dlon) ** 2
        distance = 2 * 6371.007176 * xr.ufuncs.arcsin(xr.ufuncs.sqrt(distance))

        return distance

    def get_subset_as_xarray(self, var: str, points_x: slice, points_y: slice, line_length: int = None,
                             time_counter: int = 0):
        """
        This method gets a subset of the data across the x/y indices given for the chosen variable.

        Setting time_counter to None will treat `var` as only having 3 dimensions depth, y, x

        there is a check on `var` to see the size of the time_counter, if 1 then time_counter is fixed to index 0.

        :param var: the name of the variable to get data from
        :param points_x: a list/array of indices for the x dimension
        :param points_y: a list/array of indices for the y dimension
        :param line_length: (Optional) the length of your subset (assuming simple line transect)
        :param time_counter: (Optional) which time slice to get data from, if None and the variable only has one a time
                             channel of length 1 then time_counter is fixed too an index of 0
        :return: data across all depths for the chosen variable along the given indices
        """

        try:
            [time_size, _, _, _] = self.dataset[var].shape
            if time_size == 1:
                time_counter == 0

        except ValueError:
            time_counter = None

        dx = xr.DataArray(points_x)
        dy = xr.DataArray(points_y)

        if time_counter is None:
            smaller = self.dataset[var].isel(x_dim=dx, y_dim=dy)
        else:
            smaller = self.dataset[var].isel(t_dim=time_counter, x_dim=dx, y_dim=dy)

        return smaller

    def get_2d_subset_as_xarray(self, var: str, points_x: slice, points_y: slice, line_length: int = None,
                                time_counter: int = 0):
        """

        :param var:
        :param points_x:
        :param points_y:
        :param line_length:
        :param time_counter:
        :return:
        """

        try:
            [time_size, _, _, _] = self.dataset[var].shape
            if time_size == 1:
                time_counter == 0
        except ValueError:
            time_counter = None

        if time_counter is None:
            smaller = self.dataset[var].isel(x=points_x, y=points_y)
        else:
            smaller = self.dataset[var].isel(time_counter=time_counter, x=points_x, y=points_y)

        return smaller

    def plot_simple_2d(self, x, y, data: xr.DataArray, cmap, plot_info: dict):
        """
        This is a simple method that will plot data in a 2d. It is a wrapper for matplotlib's 'pcolormesh' method.

        `cmap` and `plot_info` are required to run this method, `cmap` is passed directly to `pcolormesh`.

        `plot_info` contains all the required information for setting the figure;
         - ylim
         - xlim
         - clim
         - title
         - fig_size
         - ylabel

        :param x: The variable contain the x axis information
        :param y: The variable contain the y axis information
        :param data: the DataArray a user wishes to plot
        :param cmap:
        :param plot_info:
        :return:
        """
        import matplotlib.pyplot as plt

        plt.close('all')

        fig = plt.figure()
        plt.rcParams['figure.figsize'] = plot_info['fig_size']

        ax = fig.add_subplot(411)
        plt.pcolormesh(x, y, data, cmap=cmap)

        plt.ylim(plot_info['ylim'])
        plt.xlim(plot_info['xlim'])
        plt.title(plot_info['title'])
        plt.ylabel(plot_info['ylabel'])
        plt.clim(plot_info['clim'])
        plt.colorbar()

        return plt

    def plot_cartopy(self, var: str, plot_var: array, params, time_counter: int = 0):
        try:
            import cartopy.crs as ccrs  # mapping plots
            import cartopy.feature  # add rivers, regional boundaries etc
            from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER  # deg symb
            from cartopy.feature import NaturalEarthFeature  # fine resolution coastline
        except ImportError:
            import sys
            warn("No cartopy found - please run\nconda install -c conda-forge cartopy")
            sys.exit(-1)

        import matplotlib.pyplot as plt
        plt.close('all')
        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca()
        ax = plt.subplot(1, 1, 1, projection=ccrs.PlateCarree())

        cset = self.dataset[var].isel(time_counter=time_counter, deptht=0).plot.pcolormesh \
            (np.ma.masked_where(plot_var == np.NaN, plot_var), transform=ccrs.PlateCarree(), cmap=params.cmap)

        cset.set_clim([params.levs[0], params.levs[-1]])

        ax.add_feature(cartopy.feature.OCEAN)
        ax.add_feature(cartopy.feature.BORDERS, linestyle=':')
        ax.add_feature(cartopy.feature.RIVERS)
        coast = NaturalEarthFeature(category='physical', scale='10m', facecolor='none', name='coastline')
        ax.add_feature(coast, edgecolor='gray')

        gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
                          linewidth=0.5, color='gray', alpha=0.5, linestyle='-')

        gl.xlabels_top = False
        gl.xlabels_bottom = True
        gl.ylabels_right = False
        gl.ylabels_left = True
        gl.xformatter = LONGITUDE_FORMATTER
        gl.yformatter = LATITUDE_FORMATTER

        plt.colorbar(cset, shrink=params.colorbar_shrink, pad=.05)

        # tmp = self.dataset.votemper
        # tmp.attrs = self.dataset.votemper.attrs
        # tmp.isel(time_counter=time_counter, deptht=0).plot.contourf(ax=ax, transform=ccrs.PlateCarree())
        # ax.set_global()
        # ax.coastlines()
        plt.show()

    def plot_movie(self):
        raise NotImplementedError



import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import datetime
import pandas as pd
from os import listdir
from warnings import warn
from .OBSERVATION import OBSERVATION

class TIDEGAUGE(OBSERVATION):
    '''
    An object for reading, storing and manipulating tide gauge data.
    Reading and organisation methods are centred around the GESLA database.
    However, any fixed time series data can be used if in the correct format.
    (Source: https://www.gesla.org/)
    The data format used for this object is as follows:
        
    *Data Format Overview*
        
        1. Data for each fixed location is stored inside its own xarray
           Dataset object. This has one dimension: time. It can contain
           any number of other variables. For GESLA, this is sea_level and
           qc_flags. Attributes of the dataset include longitude, latitude and
           site_name
        2. Multiple locations are stored in an ordered list (of Datasets). A 
           corresponding ordered list of site names should also be contained
           within the object for quick access to specific time series.
           
    *Methods Overview*
    
        1. __init__(): Can be initialised with a GESLA directory or empty.
        2. get_gesla_filenames(): Gets the names of all GESLA files in a
           directory.
        3. read_gesla_to_xarray_v3(): Reads a format version 3.0 GESLA file to
           an xarray Dataset.
        4. read_gesla_header_v3(): Reads the header of a version 3 GESLA file.
        5. read_gesla_data_v3(): Reads data from a version 3 GESLA file.
        6. plot_map(): Plots locations of all time series on a map.
        7. plot_timeseries(): Plots a specified time series.
        8. obs_operator(): Interpolates model data to time series locations
           and times (not yet implemented).
    '''  
    
    def __init__(self, directory=None, file_list = None, 
                 date_start=None, date_end=None):
        '''
        Initialise TIDEGAUGE object either as empty (no arguments) or by
        reading GESLA data from a directory between two datetime objects.
        
        Example usage:
        --------------
        # Read all data in directory in January 1990
        date0 = datetime.datetime(1990,1,1)
        date1 = datetime.datetime(1990,2,1)
        tg = coast.TIDEGAUGE('gesla_directory/', date0, date1)

        Parameters
        ----------
        directory (str) : Path to directory containing desired GESLA files
        file_list (list of str) : list of filenames to read from directory.
                                  Optional.
        date_start (datetime) : Start date for data read. Optional
        date_end (datetime) : end date for data read. Optional

        Returns
        -------
        Self
        '''
        if type(file_list) is str:
            file_list = [file_list]
        
        # If no file list is supplied read all from directory
        if type(directory) is str and file_list is None:
            self.dataset_list=[]
            file_list, lats, lons, names = self.get_gesla_filenames(directory)
            self.latitude = lats
            self.longitude = lons
            self.site_name = names
            for ff in file_list:
                self.dataset_list.append( self.read_gesla_to_xarray_v3(ff,
                                                        date_start, date_end) )
        # If file list is supplied, read files from directory
        elif type(directory) is str and type(file_list) is list:
            self.dataset_list=[]
            self.latitude = []
            self.longitude = []
            self.site_name = []
            for ff in file_list:
                tmp_dataset = self.read_gesla_to_xarray_v3(directory + '/' + ff, 
                                                           date_start, date_end)
                self.dataset_list.append( tmp_dataset )
                self.latitude.append(tmp_dataset.attrs['latitude'])
                self.longitude.append(tmp_dataset.attrs['longitude'])
                self.site_name.append(tmp_dataset.attrs['site_name'])
        else:
            self.dataset_list = []
            self.latitude = []
            self.longitude = []
            self.site_name = []
        return
    
    def get_gesla_filenames(self, directory):
        '''
        Get all filenames in a directory. Try except is used to try and
        ensure that each file is indeed a GESLA file.
        
        Example usage:
        --------------
        file_list = TIDEGAUGE.get_gesla_filenames('<directory>')

        Parameters
        ----------
        directory (str) : Path to directory containing desired GESLA files

        Returns
        -------
        list of filenames (str), latitude (float), longitude (float) and 
        site names (str)
        '''
        file_list = listdir(directory)
        new_file_list = []
        latitude_list = []
        longitude_list = []
        sitename_list = []
        
        for ff in file_list:
            try:
                header_dict = self.read_gesla_header_v3(directory+ff)
                latitude_list.append(header_dict['latitude'])
                longitude_list.append(header_dict['longitude'])
                sitename_list.append(header_dict['site_name'])
                new_file_list.append(directory+ff)
            except:
                pass
            
        file_list = new_file_list
        latitude_list = np.array(latitude_list)
        longitude_list = np.array(longitude_list)
        
        return file_list, latitude_list, longitude_list, sitename_list
    
    def read_gesla_to_xarray_v3(self, fn_gesla, date_start=None, date_end=None):
        '''
        For reading from a single GESLA2 (Format version 3.0) file into an
        xarray dataset. Formatting according to Woodworth et al. (2017).
        Website: https://www.gesla.org/
        If no data lies between the specified dates, a dataset is still created
        containing information on the tide gauge, but the time dimension will
        be empty.
        Parameters
        ----------
        fn_gesla (str) : path to gesla tide gauge file
        date_start (datetime) : start date for returning data
        date_end (datetime) : end date for returning data
       
        Returns
        -------
        xarray.Dataset object.
        '''
        try:
            header_dict = self.read_gesla_header_v3(fn_gesla)
            dataset = self.read_gesla_data_v3(fn_gesla, date_start, date_end)
        except:
            raise Exception('Problem reading GESLA file: ' + fn_gesla)
        # Attributes
        dataset.attrs = header_dict
        
        return dataset
    
    def read_gesla_header_v3(self, fn_gesla):
        '''
        Reads header from a GESLA file (format version 3.0).
        
        Parameters
        ----------
        fn_gesla (str) : path to gesla tide gauge file
       
        Returns
        -------
        dictionary of attributes
        '''
        fid = open(fn_gesla)
        
        # Read lines one by one (hopefully formatting is consistent)
        fid.readline() # Skip first line
        # Geographical stuff
        site_name = fid.readline().split()[3:]
        site_name = '_'.join(site_name)
        country = fid.readline().split()[2:]
        country = '_'.join(country)
        contributor = fid.readline().split()[2:]
        contributor = '_'.join(contributor)
        # Coordinates
        latitude = float(fid.readline().split()[2])
        longitude = float(fid.readline().split()[2])
        coordinate_system = fid.readline().split()[3]
        # Dates
        start_date = fid.readline().split()[3:5] 
        start_date = ' '.join(start_date)
        start_date = pd.to_datetime(start_date)
        end_date = fid.readline().split()[3:5]
        end_date = ' '.join(end_date)
        end_date = pd.to_datetime(end_date)
        time_zone_hours = float(fid.readline().split()[4])
        # Other
        fid.readline() #Datum
        fid.readline() #Instrument
        precision = float(fid.readline().split()[2])
        null_value = float( fid.readline().split()[3])

        fid.close()
        # Put all header info into an attributes dictionary
        header_dict = {'site_name' : site_name, 'country':country, 
                       'contributor':contributor, 'latitude':latitude,
                       'longitude':longitude, 
                       'coordinate_system':coordinate_system,
                       'original_start_date':start_date, 
                       'original_end_date': end_date,
                       'time_zone_hours':time_zone_hours, 
                       'precision':precision, 'null_value':null_value}
        return header_dict
    
    def read_gesla_data_v3(self, fn_gesla, date_start=None, date_end=None,
                           header_length:int=32):
        '''
        Reads observation data from a GESLA file (format version 3.0).
        
        Parameters
        ----------
        fn_gesla (str) : path to gesla tide gauge file
        date_start (datetime) : start date for returning data
        date_end (datetime) : end date for returning data
        header_length (int) : number of lines in header (to skip when reading)
       
        Returns
        -------
        xarray.Dataset containing times, sealevel and quality control flags
        '''
        # Initialise empty dataset and lists
        dataset = xr.Dataset()
        time = []
        sea_level = []
        qc_flags = []
        # Open file and loop until EOF
        with open(fn_gesla) as file:
            line_count = 0
            for line in file:
                # Read all data. Date boundaries are set later.
                if line_count>header_length:
                    working_line = line.split()
                    if working_line[0] != '#':
                        time.append(working_line[0] + ' ' + working_line[1])
                        sea_level.append(float(working_line[2]))
                        qc_flags.append(int(working_line[3]))
                    
                line_count = line_count + 1
             
        # Convert time list to datetimes using pandas
        time = np.array(pd.to_datetime(time))
        
        # Return only values between stated dates
        start_index = 0
        end_index = len(time)
        if date_start is not None:
            date_start = np.datetime64(date_start)
            start_index = np.argmax(time>=date_start)
        if date_end is not None:
            date_end = np.datetime64(date_end)
            end_index = np.argmax(time>date_end)
        time = time[start_index:end_index]
        sea_level = sea_level[start_index:end_index]
        qc_flags=qc_flags[start_index:end_index]
        
        # Set null values to nan
        sea_level = np.array(sea_level)
        qc_flags = np.array(qc_flags)
        sea_level[qc_flags==5] = np.nan
        
        # Assign arrays to Dataset
        dataset['time'] = xr.DataArray(time, dims=['time'])
        dataset['sea_level'] = xr.DataArray(sea_level, dims=['time'])
        dataset['qc_flags'] = xr.DataArray(qc_flags, dims=['time'])
        
        # Assign local dataset to object-scope dataset
        return dataset
    
    def plot_map(self):
        '''
        Plot tide gauge locations on a map
        
        Example usage:
        --------------
        # For a TIDEGAUGE object tg
        tg.plot_map()

        '''
        try:
            import cartopy.crs as ccrs  # mapping plots
            import cartopy.feature  # add rivers, regional boundaries etc
            from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER  # deg symb
            from cartopy.feature import NaturalEarthFeature  # fine resolution coastline
        except ImportError:
            import sys
            warn("No cartopy found - please run\nconda install -c conda-forge cartopy")
            sys.exit(-1)

        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(10, 10))
        ax = fig.gca()
        ax = plt.subplot(1, 1, 1, projection=ccrs.PlateCarree())

        cset = plt.scatter(self.longitude, self.latitude, c='k')

        ax.add_feature(cartopy.feature.BORDERS, linestyle=':')
        coast = NaturalEarthFeature(category='physical', scale='50m',
                                    facecolor=[0.8,0.8,0.8], name='coastline',
                                    alpha=0.5)
        ax.add_feature(coast, edgecolor='gray')
        plt.title('Map of gauge locations')
        plt.ylabel('Latitude')
        plt.xlabel('Longitude')
        plt.show()

        gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True,
                          linewidth=0.5, color='gray', alpha=0.5, linestyle='-')

        gl.top_labels = False
        gl.bottom_labels = True
        gl.right_labels = False
        gl.left_labels = True
        gl.xformatter = LONGITUDE_FORMATTER
        gl.yformatter = LATITUDE_FORMATTER
        
        return fig, ax
    
    def plot_timeseries(self, site, date_start=None, date_end=None, 
                        var_name = 'sea_level', qc_colors=True, 
                        plot_line = False):
        '''
        Quick plot of time series stored within object's dataset
        Parameters
        ----------
        site (str or int) : Either site name as a string or site index (int)
                            inside dataset list.
        date_start (datetime) : Start date for plotting
        date_end (datetime) : End date for plotting
        var_name (str) : Variable to plot. Default: sea_level
        qc_colors (bool) : If true, markers are coloured according to qc values
        plot_line (bool) : If true, draw line between markers
       
        Returns
        -------
        matplotlib figure and axes objects
        '''
        
        if type(site) is int:
            dataset = self.dataset_list[site]
            site_name = dataset.site_name
            # Numpyify data
            x = np.array(dataset.time)
            y = np.array(dataset[var_name])
            qc = np.array(dataset.qc_flags)
        elif type(site) is str:
            index = self.site_name.index(site)
            dataset = self.dataset_list[index]
            site_name = dataset.site_name
            # Numpyify data
            x = np.array(dataset.time)
            y = np.array(dataset.sea_level)
            qc = np.array(dataset.qc_flags)
        else:
            raise Exception('site argument for plot_timeseries_single' + 
                            ' must be int or str.')
        
        # Use only values between stated dates
        start_index = 0
        end_index = len(x)
        if date_start is not None:
            date_start = np.datetime64(date_start)
            start_index = np.argmax(x>=date_start)
        if date_end is not None:
            date_end = np.datetime64(date_end)
            end_index = np.argmax(x>date_end)
        x = x[start_index:end_index]
        y = y[start_index:end_index]
        qc = qc[start_index:end_index]
        
        # Plot lines first if needed
        if plot_line:
            plt.plot(x,y, c=[0.5,0.5,0.5], linestyle='--', linewidth=0.5)
        
        # Two plotting routines for whether or not to use qc flags.
        if qc_colors:
            size = 5
            fig = plt.figure(figsize=(10,10))
            ax = plt.scatter(x[qc==0], y[qc==0], s=size)
            plt.scatter(x[qc==1], y[qc==1], s=size)
            plt.scatter(x[qc==2], y[qc==2], s=size)
            plt.scatter(x[qc==3], y[qc==3], s=size)
            plt.scatter(x[qc==4], y[qc==4], s=size)
            plt.grid()
            plt.legend(['No QC','Correct','Interpolated','Doubtful','Spike'],
                       loc='upper left', ncol=5)
            plt.xticks(rotation=45)
        else:
            fig = plt.figure(figsize=(10,10))
            plt.scatter(x,y)
            plt.grid()
            plt.xticks(rotation=65)
            
        # Title and axes
        plt.xlabel('Date')
        plt.ylabel(var_name + ' (m)')
        plt.title(var_name + ' at site: ' + site_name)
        
        return fig, ax

    def obs_operator():
        return